import threading
import queue

from .lib_test_utils import *

from warawara import *

import warawara
stream = warawara.subproc.stream


def queue_to_list(Q):
    ret = []
    while not Q.empty():
        ret.append(Q.get())
    return ret


class TestEventBroadcaster(TestCase):
    def test_all(self):
        data1 = []
        def handler1(arg):
            data1.append(arg)

        data2 = []
        def handler2(arg):
            data2.append(arg)

        import warawara
        hub = warawara.subproc.EventBroadcaster()
        hub.broadcast('...')

        hub += handler1
        hub.broadcast('wah')

        hub += handler2
        hub.broadcast('Wah')

        hub += handler2
        hub.broadcast('WAAAAAH')

        hub -= handler1
        hub.broadcast('wah?')

        hub -= handler2
        hub.broadcast('wow')

        hub -= handler2
        hub.broadcast('bye')

        self.eq(data1, ['wah', 'Wah', 'WAAAAAH'])
        self.eq(data2, ['Wah', 'WAAAAAH', 'WAAAAAH', 'wah?', 'wah?', 'wow'])


class TestStream(TestCase):
    def test_stream_basic_io(self):
        s = stream()
        self.eq(s.keep, False)

        s.writeline('line1')
        s.writeline('line2')
        s.writeline('line3')
        self.false(s.closed)

        s.close()
        self.true(s.closed)

        self.eq(s.lines, [])
        self.eq(s.readline(), 'line1')
        self.eq(s.readline(), 'line2')
        self.eq(s.readline(), 'line3')

    def test_stream_iter(self):
        s = stream()
        self.eq(s.keep, False)
        s.writelines(['line1', 'line2', 'line3'])

        i = iter(s)
        self.eq(next(i), 'line1')
        self.eq(next(i), 'line2')
        self.eq(next(i), 'line3')

        s.close()

        with self.raises(StopIteration):
            next(iter(s))

    def test_stream_keep(self):
        s = stream()
        s.keep = True

        lines = ['line1', 'line2', 'line3']
        for line in lines:
            s.writeline(line)

        self.eq(s.lines, lines)
        self.eq(len(s), 3)

        self.false(s.empty)
        self.true(bool(s))

        self.eq(s.readline(), 'line1')
        self.eq(s.readline(), 'line2')
        self.eq(s.readline(), 'line3')

        self.eq(s.lines, lines)
        self.eq(len(s), 3)

    def test_stream_subscribers(self):
        data1 = []
        def handler1(line):
            data1.append(line)

        data2 = []
        def handler2(line):
            data2.append(line)

        Q = queue.Queue()

        s = stream()
        s.welcome([handler1, handler2])
        s.welcome(Q)
        s.welcome(True)

        with self.raises(TypeError):
            s.welcome(s)

        lines = ['line1', 'line2', 'line3']
        s.writelines(lines)

        self.eq(data1, lines)
        self.eq(data2, lines)

    def test_stream_write_after_close(self):
        def should_not_be_called_handler(line):
            self.fail()

        # test in test
        with self.raises(AssertionError):
            should_not_be_called_handler('wah')

        s = stream()
        s.welcome(should_not_be_called_handler)
        s.close()

        s.writeline('line1')

        with self.raises(BrokenPipeError):
            s.writeline('line2', suppress=False)


class TestSubproc(TestCase):
    def test_default_properties(self):
        def prog(proc):
            self.eq(proc[0].readline(), 'line')

        p = run(prog, stdin='line', stdout=False, stderr=False)

    def test_stdout(self):
        p = run('seq 5'.split())
        self.eq(p.stdout.lines, '1 2 3 4 5'.split())

    def test_disable_stdout_and_stderr(self):
        p = command('seq 5'.split(), stdout=None, stderr=None)
        self.true(p.stdout.closed)
        self.true(p.stderr.closed)

    def test_wait_early(self):
        p = command('seq 5'.split())
        p.wait()
        p.run(wait=False)
        p.wait()
        self.eq(p.stdout.lines, '1 2 3 4 5'.split())

    def test_already_running_error(self):
        checkpoint = self.checkpoint()

        def prog(proc, *args):
            checkpoint.wait()

        p = command(prog)
        p.run(wait=False)

        with self.raises(AlreadyRunningError):
            p.run()

        checkpoint.set()
        p.wait()

        p = command(['sleep', 1])
        p.run(wait=False)
        with self.raises(AlreadyRunningError):
            p.run(wait=False)
        p.kill()

    def test_word(self):
        p = run('true')
        self.eq(p.returncode, 0)

        p = run('false')
        self.eq(p.returncode, 1)

    def test_invalid_cmd(self):
        with self.raises(ValueError):
            p = command()
        with self.raises(ValueError):
            p = run()

        for i in ([], True, 3, None, queue.Queue()):
            with self.raises(ValueError):
                command(i)
            with self.raises(ValueError):
                run(i)

    def test_run_with_context_manager(self):
        barrier = threading.Barrier(2)

        def prog(proc, *args):
            barrier.wait()

        with command(prog) as p:
            barrier.wait()

    def test_stdout_nokeep(self):
        p = command('seq 5'.split(), stdout=False)
        p.run()
        self.eq(p.stdout.lines, [])

    def test_keep_trailing_whitespaces(self):
        p = run(['echo', 'a b c '])
        self.eq(p.stdout.lines, ['a b c '])

    def test_callable(self):
        def prog(proc, *args):
            self.eq(args, ('arg1', 'arg2'))
            for idx, line in enumerate(proc[0]):
                proc[(idx % 2) + 1].writeline(line)
            return 2024

        p = run([prog, 'arg1', 'arg2'], stdin=['hello ', 'how are you ', 'im fine ', 'thank you '])
        self.eq(p.stdout.lines, ['hello ', 'im fine '])
        self.eq(p.stderr.lines, ['how are you ', 'thank you '])
        self.eq(p.returncode, 2024)

    def test_stdout_callback(self):
        lines = []
        def callback(line):
            lines.append(line)
        p = command('seq 5'.split(), stdout=callback)
        p.run()
        self.eq(lines, ['1', '2', '3', '4', '5'])

    def test_stdout_queue(self):
        Q = queue.Queue()
        p = command('seq 5'.split(), stdout=Q)
        self.false(p.stdout.keep)
        p.run()
        self.eq(p.stdout.lines, [])
        self.eq(queue_to_list(Q), ['1', '2', '3', '4', '5'])

    def test_multi_ouptut_merge(self):
        def prog(proc, *args):
            for i in range(5):
                proc[1].write(i)
                proc[2].write(i)

        Q = queue.Queue()

        lines = []
        def callback(line):
            lines.append(line)

        self.true(hasattr(Q, 'put'))

        p = command(prog, stdout=(Q, callback, True), stderr=(Q, True))
        p.run()
        self.eq(p.stdout.lines, [0, 1, 2, 3, 4])
        self.eq(p.stderr.lines, [0, 1, 2, 3, 4])
        self.eq(lines, [0, 1, 2, 3, 4])

        self.eq(queue_to_list(Q), [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])

    def test_stdin(self):
        p = command('nl -w 1 -s :'.split(), stdin=['hello', 'world'])
        p.run()
        self.eq(p.stdout.lines, ['1:hello', '2:world'])

    def test_stdin_delayed_write(self):
        data = ['hello', 'world']
        p = command('nl -w 1 -s :'.split(), stdin=data)

        data += ['wow', 'haha']     # ignored
        p.stdin.writeline('wah')    # allow write before run
        p.run(wait=False).wait()

        self.eq(p.stdout.lines, ['1:hello', '2:world', '3:wah'])

    def test_stdin_queue(self):
        Q = queue.Queue()
        p = command('nl -w 1 -s :'.split(), stdin=Q)

        Q.put('pre')

        p.run(wait=False)

        Q.join()
        Q.put('hello')
        Q.put('world')
        Q.join()
        Q.put('wah')
        Q.join()

        p.stdin.close()

        p.wait()
        self.eq(p.stdout.lines, ['1:pre', '2:hello', '3:world', '4:wah'])

    def test_callable_with_pipe(self):
        def prog(proc, *args):
            for line in proc[0]:
                proc[1].writeline(line)
                proc[2].writeline(line)
            return 2024

        p1 = command(prog, stdin=['hello', 'world'])
        p2 = command('nl -w 1 -s :'.split(), stdin=True)
        p3 = command('nl -w 1 -s /'.split(), stdin=True)
        pipe(p1.stdout, p2.stdin)
        pipe(p1.stderr, p3.stdin)

        p1.run()
        p2.run()
        p3.run()

        self.eq(p1.returncode, 2024)
        self.eq(p2.stdout.lines, ['1:hello', '2:world'])
        self.eq(p3.stdout.lines, ['1/hello', '2/world'])

    def test_callable_raises_exception(self):
        def prog(proc, *args):
            # NameError
            n + 1

        with self.raises(NameError):
            p = run(prog)

    def test_loopback(self):
        # Collatz function
        def collatz_function(x):
            return (x // 2) if (x % 2 == 0) else (3 * x + 1)

        def prog(proc, *args):
            for line in proc[0]:
                x = line
                if x == 1:
                    break
                else:
                    proc[1].write(collatz_function(x))

        p = command(prog, stdin=True)

        # loopback
        pipe(p.stdout, p.stdin)

        import time
        t = int(time.time())
        p.stdin.write(t)
        p.run()

        # If this test fails, make sure to check the initial input
        self.eq(p.stdout.lines[-1], 1, p.stdin.lines[0])

    def test_wait_false(self):
        checkpoint = self.checkpoint()

        def prog(proc, *args):
            checkpoint.wait()

        p = command(prog)
        p.run(wait=False)
        checkpoint.check(False)
        p.wait(False)
        checkpoint.check(False)
        checkpoint.set()

    def test_wait_invalid_types(self):
        def prog(proc, *args): # pragma: no cover
            pass

        p = command(prog)
        with self.raises(TypeError):
            p.run(wait='wah')
        self.eq(p.proc, None)
        self.eq(p.thread, None)

    def test_timeout(self):
        import time

        p = command(['sleep', 3])
        t1 = time.time()
        try:
            p.run(wait=0.1)
        except TimeoutExpired:
            pass
        t2 = time.time()
        self.le(t2 - t1, 1)
        p.kill()

    def test_poll(self):
        p = command('true')
        self.eq(p.poll(), False)
        p.run()
        self.eq(p.poll(), 0)

        checkpoint = self.checkpoint()
        def prog(proc, *args):
            checkpoint.wait()
            return 1
        p = command(prog)
        self.eq(p.poll(), False)
        p.run(wait=False)
        self.eq(p.poll(), None)
        checkpoint.set()
        p.wait()
        self.eq(p.poll(), 1)

    def test_signal(self):
        p = run(['cat'], wait=False)
        import signal
        p.signal(signal.SIGINT)
        p.wait()
        self.eq(p.signaled, signal.SIGINT)

    def test_kill_callable(self):
        checkpoint = self.checkpoint()

        def prog(proc, *args):
            proc.killed.wait()
            checkpoint.set()

        p = run(prog, wait=False)
        self.eq(p.signaled, False)
        self.eq(p.signaled, None)
        p.kill()
        p.wait()

        import signal
        checkpoint.check()
        self.eq(p.signaled, signal.SIGKILL)

        p.signaled.clear()
        self.eq(p.signaled, False)
        self.eq(p.signaled, None)

        p.kill(signal.SIGTERM)
        self.eq(p.signaled, signal.SIGTERM)

    def test_read_stdout_twice(self):
        ans = '1 2 3 4 5'.split()
        p = run('seq 5'.split())

        lines = []
        for line in p.stdout:
            lines.append(line)
        self.eq(lines, ans)

        checkpoint = self.checkpoint()
        lines = []
        def may_stuck():
            for line in p.stdout:
                lines.append(line)
            checkpoint.set()

        with self.run_in_thread(may_stuck):
            checkpoint.check()
            self.eq(lines, ans)

    def test_encoding_false(self):
        pi = b''
        pi += b'\x31\x41\x59\x26\x53\x58\x97\x93\x23\x84\x62\x64\x33\x83\x27\x95'
        pi += b'\x02\x88\x41\x97\x16\x93\x99\x37\x51\x05\x82\x09\x74\x94\x45\x92'
        pi += b'\x30\x78\x16\x40\x62\x86\x20\x89\x98\x62\x80\x34\x82\x53\x42\x11'
        pi += b'\x70\x67\x90'
        p = run(['xxd'], encoding=False, stdin=pi)
        self.eq(p.returncode, 0)
        self.eq(p.stdin.lines, [pi])
        self.eq(len(p.stdout.lines), 1)
        hex_pi = p.stdout.lines[0].decode('utf-8')
        self.eq(hex_pi, '''
00000000: 3141 5926 5358 9793 2384 6264 3383 2795  1AY&SX..#.bd3.'.
00000010: 0288 4197 1693 9937 5105 8209 7494 4592  ..A....7Q...t.E.
00000020: 3078 1640 6286 2089 9862 8034 8253 4211  0x.@b. ..b.4.SB.
00000030: 7067 90                                  pg.
'''.lstrip())

    def test_encoding_true_but_write_binary(self):
        pi = b''
        pi += b'\x31\x41\x59\x26\x53\x58\x97\x93\x23\x84\x62\x64\x33\x83\x27\x95'
        pi += b'\x02\x88\x41\x97\x16\x93\x99\x37\x51\x05\x82\x09\x74\x94\x45\x92'
        pi += b'\x30\x78\x16\x40\x62\x86\x20\x89\x98\x62\x80\x34\x82\x53\x42\x11'
        pi += b'\x70\x67\x90'
        p = run(['xxd'], stdin=pi)
        self.eq(p.returncode, 0)
        self.eq(p.stdin.lines, [pi])
        self.eq(len(p.stdout.lines), 4)
        self.eq(p.stdout.lines, [
            "00000000: 3141 5926 5358 9793 2384 6264 3383 2795  1AY&SX..#.bd3.'.",
            "00000010: 0288 4197 1693 9937 5105 8209 7494 4592  ..A....7Q...t.E.",
            "00000020: 3078 1640 6286 2089 9862 8034 8253 4211  0x.@b. ..b.4.SB.",
            "00000030: 7067 90                                  pg.",
            ])

    def test_large_amount_of_retention_data(self):
        p = command(['echo', 'a lot of data'], encoding=False)
        def mock_poll():
            # block p.poll() until the p.proc writes all stdout and exits
            while p.proc.poll() is None:
                pass
            return p.proc.poll()

        p.poll = mock_poll
        p.run()
        self.eq(p.stdout.lines, [b'a lot of data\n'])


class TestPipe(TestCase):
    def test_pipe(self):
        p1 = command('nl -w 1 -s :'.split(), stdin=['hello', 'world'])
        p2 = command('nl -w 1 -s /'.split(), stdin=True)
        pp = pipe(p1.stdout, p2.stdin)

        p1.run()
        self.eq(p1.stdout.lines, ['1:hello', '2:world'])
        p2.run(wait=False)

        p2.wait()
        self.eq(p2.stdin.lines, ['1:hello', '2:world'])
        self.eq(p2.stdout.lines, ['1/1:hello', '2/2:world'])

        pp.join()

    def test_pipe_dont_start(self):
        p1 = command(lambda prog: prog.stdout.writeline('wah'))

        def cat(prog):
            for line in prog.stdin:
                prog.stdout.writeline(line)
        p2 = command(cat, stdin=True)

        pp = pipe(p1.stdout, p2.stdin, start=False)

        p1.run(wait=False)
        p2.run(wait=False)
        p1.wait()

        self.false(p2.stdout.lines)
        self.eq(p2.returncode, None)

        pp.start()
        p2.wait()
        pp.join()

        self.eq(p2.stdout.lines, ['wah'])

    def test_pipe_istream_already_closed(self):
        i = stream()
        o = stream()
        i.close()

        with self.raises(EOFError):
            pipe(i, o)

    def test_pipe_ostream_already_closed(self):
        i = stream()
        o = stream()
        o.close()

        with self.raises(BrokenPipeError):
            pipe(i, o)

    def test_pipe_exception(self):
        i = stream()
        o = stream()
        o.queue = None
        o.close = lambda: None

        p = pipe(i, o)
        i.writeline('wah')

        with self.raises(Exception):
            p.join()

        self.ne(p.exception, None)

    def test_pipe_merge_to_same_stream(self):
        i1 = stream()
        i2 = stream()
        o = stream()
        o.keep = True

        pp1 = pipe(i1, o)
        e1 = threading.Event()
        pp1.post_write = e1.set
        pp2 = pipe(i2, o)
        e2 = threading.Event()
        pp2.post_write = e2.set

        i1.writeline('wah1')
        e1.wait()
        e1.clear()
        self.eq(o.lines, ['wah1'])

        i2.writeline('wow1')
        e2.wait()
        e2.clear()
        self.eq(o.lines, ['wah1', 'wow1'])

        i1.writeline('wah2')
        e1.wait()
        e1.clear()
        self.eq(o.lines, ['wah1', 'wow1', 'wah2'])

        i1.close()
        pp1.join()
        self.false(o.closed)

        i2.writeline('wow2')
        e2.wait()
        e2.clear()
        self.eq(o.lines, ['wah1', 'wow1', 'wah2', 'wow2'])

        i2.close()
        pp2.join()

        self.true(o.closed)



from .lib_test_utils import *

from warawara import *


class TestRegex(TestCase):
    def test_match(self):
        rec = rere('wara wa ra')

        m = rec.match(r'^(\w+) (\w+)$')
        self.eq(m, None)

        m = rec.match(r'^(\w+) (\w+) (\w+)$')
        self.eq(m.groups(), ('wara', 'wa', 'ra'))
        self.eq(m.group(2), 'wa')

        self.eq(rec.groups(), m.groups())
        self.eq(rec.group(2), m.group(2))

        self.eq(rec.sub(r'wara', 'WARA'), 'WARA wa ra')

        rec.search(r'\bwa\b')
        self.eq(rec.start(), 5)
        self.eq(rec.end(), 7)

        self.ne(rec.fullmatch(r'(\w+) (\w+) (\w+)'), None)

        self.eq(rec.split(' +'), rec.text.split(' '))

        self.eq(rec.findall(r'\b\w+\b'), ['wara', 'wa', 'ra'])

        self.eq(rec.subn(r'wa', 'WA'), ('WAra WA ra', 2))



import os
from pathlib import Path

from .internal_utils import exporter
export, __all__ = exporter()


@export
def cwd(path=None):
    try:
        if path is None:
            here = Path.cwd()
            return here

        os.chdir(path)
        return Path.cwd()

    except:
        return False


dir_stack = []

@export
class pushd:
    def __init__(self, path=None):
        here = cwd()
        if not here:
            self.succ = False

        else:
            self.succ = cwd(path)
            if self.succ:
                dir_stack.append(here)

    def __bool__(self):
        return self.succ

    def __enter__(self):
        return self.succ

    def __exit__(self, exc_type, exc_value, traceback):
        popd()


@export
def popd(all=False):
    if not dir_stack:
        return False

    try:
        if all:
            os.chdir(dir_stack[0])
            dir_stack.clear()

        else:
            os.chdir(dir_stack.pop())

        return True

    except:
        pass

    return False


@export
def dirs(clear=False):
    if clear:
        dir_stack.clear()
    return list(dir_stack) + [cwd()]


@export
def home():
    return Path.home()


@export
def shrinkuser(path):
    path = str(path)
    trailing_slash = '/' if path.endswith('/') else ''

    import os.path
    path = path.rstrip('/') + '/'
    homepath = os.path.expanduser('~').rstrip('/') + '/'
    if path.startswith(homepath):
        ret = os.path.join('~', path[len(homepath):])
    else:
        ret = path
    return ret.rstrip('/') + trailing_slash



import queue
import subprocess as sub
import threading

from signal import SIGKILL

from .lib_itertools import is_iterable

from .internal_utils import exporter
export, __all__ = exporter()


export('TimeoutExpired')
TimeoutExpired = sub.TimeoutExpired


@export
class AlreadyRunningError(Exception):
    def __init__(self, cmd):
        if callable(cmd.cmd[0]):
            prog = cmd.cmd[0].__name__ + '()'
        else:
            prog = cmd.cmd[0]

        super().__init__(' '.join(
            [prog] + cmd.cmd[1:]))


class EventBroadcaster:
    def __init__(self):
        self.handlers = []

    def __iadd__(self, handler):
        self.handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.handlers.remove(handler)
        return self

    def broadcast(self, *args, **kwargs):
        for handler in self.handlers:
            handler(*args, **kwargs)


class QueueEventAdapter:
    def __init__(self, Q):
        self.Q = Q

    def __call__(self, line):
        self.Q.put(line)


class stream:
    def __init__(self):
        self.queue = queue.Queue()
        self.keep = False
        self.lines = []
        self.eof = threading.Event()
        self.hub = EventBroadcaster()

        self.pipe_count_lock = threading.Lock()
        self.pipe_count = 0

    def welcome(self, subscriber):
        if isinstance(subscriber, (list, tuple)):
            for s in subscriber:
                self.welcome_one(s)
        else:
            self.welcome_one(subscriber)

    def welcome_one(self, subscriber):
        if subscriber is True:
            self.keep = True

        else:
            handler = None
            if hasattr(subscriber, 'put'):
                handler = QueueEventAdapter(subscriber)
            elif callable(subscriber):
                handler = subscriber

            if handler:
                self.hub += handler
            else:
                raise TypeError('Invalid subscriber value: {}'.format(repr(subscriber)))

    def pipe_attached(self):
        self.pipe_count_lock.acquire()
        try:
            self.pipe_count += 1
        finally:
            self.pipe_count_lock.release()

    def pipe_detached(self):
        self.pipe_count_lock.acquire()
        try:
            self.pipe_count -= 1
            if self.pipe_count <= 0:
                self.close()
        finally:
            self.pipe_count_lock.release()

    def read(self):
        data = self.queue.get()
        return data

    def readline(self):
        return self.read()

    def write(self, data, *, suppress=True):
        if self.closed:
            if suppress:
                return
            raise BrokenPipeError('stream already closed')

        if self.keep:
            self.lines.append(data)

        self.queue.put(data)
        self.hub.broadcast(data)

    def writeline(self, line, *, suppress=True):
        self.write(line, suppress=suppress)

    def writelines(self, lines):
        for line in lines:
            self.writeline(line)

    def close(self):
        self.eof.set()
        self.queue.put(None)

    @property
    def closed(self):
        return self.eof.is_set()

    @property
    def empty(self):
        return not self.lines and self.queue.empty()

    def __bool__(self):
        return not self.empty

    def __len__(self):
        return len(self.lines)

    def __iter__(self):
        if self.closed:
            yield from self.lines

        else:
            while True:
                line = self.readline()
                if line is None:
                    break
                yield line


class IntegerEvent(threading.Event):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = None

    def set(self, value=None):
        self.value = value
        super().set()

    def clear(self):
        self.value = None
        super().clear()

    def __eq__(self, other):
        if not self.is_set():
            return other is False or other is None
        return self.value == other


@export
class command:
    def __init__(self, cmd=None, *,
                 stdin=None, stdout=True, stderr=True,
                 encoding='utf8', rstrip='\r\n',
                 bufsize=-1,
                 env=None):

        if cmd and isinstance(cmd, str):
            cmd = [cmd]
        elif callable(cmd):
            cmd = [cmd]
        elif isinstance(cmd, (tuple, list)):
            pass
        else:
            raise ValueError('Invalid command:' + repr(cmd))

        if not cmd:
            raise ValueError('command is empty')

        if callable(cmd[0]):
            self.cmd = [token for token in cmd]
        else:
            self.cmd = [str(token) for token in cmd]

        self.encoding = encoding
        self.bufsize = bufsize
        self.rstrip = rstrip

        self.env = env
        self.proc = None
        self.thread = None
        self.exception = None
        self.signaled = IntegerEvent()
        self.returncode = None

        if isinstance(stdin, (str, bytes, bytearray)):
            stdin = [stdin]

        # Initialize stdin stream
        self.stdin = stream()
        self.stdin.keep = True
        self.stdin_queue = None
        self.stdin_autoclose = False
        if stdin is None or stdin is False:
            self.proc_stdin = None
            self.stdin.close()
        else:
            self.proc_stdin = sub.PIPE
            if isinstance(stdin, queue.Queue):
                self.stdin_queue = stdin
            elif is_iterable(stdin):
                for line in stdin:
                    self.stdin.write(line)
                self.stdin_autoclose = True

        # Initialize stdout stream
        self.stdout = stream()
        if stdout is None:
            self.proc_stdout = None
            self.stdout.close()
        elif stdout is False:
            self.proc_stdout = sub.DEVNULL
            self.stdout.close()
        else:
            self.proc_stdout = sub.PIPE
            self.stdout.keep = False
            self.stdout.welcome(stdout)

        # Initialize stderr stream
        self.stderr = stream()
        if stderr is None:
            self.proc_stderr = None
            self.stderr.close()
        elif stderr is False:
            self.proc_stderr = sub.DEVNULL
            self.stderr.close()
        else:
            self.proc_stderr = sub.PIPE
            self.stderr.keep = False
            self.stderr.welcome(stderr)

        self.io_threads = []

    @property
    def killed(self):
        return self.signaled

    def __getitem__(self, idx):
        return [self.stdin, self.stdout, self.stderr][idx]

    def __enter__(self):
        return self.run(wait=False)

    def __exit__(self, exc_type, exc_value, traceback):
        self.stdin.close()
        self.stdout.close()
        self.stderr.close()
        self.wait()

    def run(self, wait=None):
        if wait is not None and not isinstance(wait, (int, bool, float)):
            raise TypeError('The type of "wait" should be NoneType, int, bool, or float')

        if self.proc or self.thread:
            raise AlreadyRunningError(self)

        if callable(self.cmd[0]):
            def worker():
                try:
                    self.returncode = self.cmd[0](self, *self.cmd[1:])
                except Exception as e:
                    self.exception = e

                self.stdin.close()
                self.stdout.close()
                self.stderr.close()

            self.thread = threading.Thread(target=worker)
            self.thread.daemon = True
            self.thread.start()

        else:
            if self.encoding == False:
                # binary mode
                kwargs = {
                        'bufsize': 2 if self.bufsize == 1 else self.bufsize,
                        'text': False,
                        }
            else:
                kwargs = {
                        'bufsize': 1,
                        'encoding': self.encoding,
                        'errors': 'backslashreplace',
                        }

            self.proc = sub.Popen(
                    self.cmd,
                    stdin=self.proc_stdin,
                    stdout=self.proc_stdout,
                    stderr=self.proc_stderr,
                    env=self.env, **kwargs)

            def writer(self_stream, proc_stream):
                for line in self_stream:
                    if self.encoding == False:
                        proc_stream.write(line)
                    elif isinstance(line, (bytes, bytearray)):
                        proc_stream.buffer.write(line)
                    else:
                        proc_stream.write(line + '\n')
                    proc_stream.flush()
                proc_stream.close()

            def reader(self_stream, proc_stream):
                if self.encoding != False:
                    # text
                    for line in proc_stream:
                        line = line.rstrip(self.rstrip)
                        self_stream.writeline(line)

                else:
                    # binary
                    while self.poll() is None:
                        data = proc_stream.read(
                                -1
                                if self.bufsize < 0
                                else (self.bufsize or 1)
                                )

                        if not data:
                            continue

                        self_stream.write(data)

                    # Read all remaining data left in stream
                    data = proc_stream.read()
                    if data:
                        self_stream.writeline(data)

                self_stream.close()
                proc_stream.close()

            for (worker, self_stream, proc_stream) in (
                    (writer, self.stdin, self.proc.stdin),
                    (reader, self.stdout, self.proc.stdout),
                    (reader, self.stderr, self.proc.stderr),
                    ):
                if self_stream is not None and proc_stream is not None:
                    t = threading.Thread(target=worker, args=(self_stream, proc_stream))
                    t.daemon = True
                    t.start()
                    self.io_threads.append(t)

        # Pull data from stdin_queue and feed into stdin stream
        if self.stdin_queue:
            def feeder():
                while True:
                    self.stdin.writeline(self.stdin_queue.get())
                    self.stdin_queue.task_done()

            t = threading.Thread(target=feeder)
            t.daemon = True
            t.start()

        elif self.stdin_autoclose:
            self.stdin.close()

        self.wait(wait)

        return self

    def poll(self):
        if self.proc:
            return self.proc.poll()
        if self.thread:
            return self.returncode
        return False

    def wait(self, timeout=None):
        if timeout is True:
            timeout = None
        elif timeout is False:
            return

        # Wait for child process to finish
        if self.proc:
            self.proc.wait(timeout)
            self.returncode = self.proc.returncode

        if self.thread:
            self.thread.join(timeout)

        # Wait too early
        if self.proc is None and self.thread is None:
            return

        if self.exception:
            raise self.exception

        # Wait for all streams to close
        self.stdin.eof.wait()
        self.stdout.eof.wait()
        self.stderr.eof.wait()

        # Gracefully wait for threads to finish
        for t in self.io_threads:
            t.join()

    def signal(self, signal):
        if self.proc:
            self.proc.send_signal(signal)

        self.signaled.set(signal)

    def kill(self, signal=SIGKILL):
        self.signal(signal)

        if self.proc:
            self.proc.wait()
            for proc_stream in (
                    self.proc.stdin,
                    self.proc.stdout,
                    self.proc.stderr
                    ):
                if proc_stream:
                    proc_stream.close()

            self.returncode = self.proc.returncode

        if self.thread:
            self.thread.join()


@export
def run(cmd=None, *,
        stdin=None, stdout=True, stderr=True,
        encoding='utf8', rstrip='\r\n',
        bufsize=-1,
        env=None,
        wait=True):
    ret = command(cmd,
                  stdin=stdin, stdout=stdout, stderr=stderr,
                  encoding=encoding,
                  rstrip=rstrip, env=env)
    ret.run(wait=wait)
    return ret


class Pipe:
    def __init__(self, istream, *ostreams):
        if istream.closed:
            raise EOFError('istream already closed')

        for ostream in ostreams:
            if ostream.closed:
                raise BrokenPipeError('ostream already closed')

        self.exception = None
        self.thread = None
        self.istream = istream
        self.ostreams = ostreams
        self.post_write = None

    def main(self):
        try:
            for line in self.istream:
                for ostream in self.ostreams:
                    ostream.write(line)
                if self.post_write:
                    self.post_write()

        except Exception as e:
            self.exception = e
            self.istream.close()

        self.istream.eof.wait()
        for ostream in self.ostreams:
            ostream.pipe_detached()

    def start(self):
        self.thread = threading.Thread(target=self.main)
        self.thread.daemon = True
        self.thread.start()

    def join(self):
        self.thread.join()
        if self.exception:
            raise self.exception


@export
def pipe(istream, *ostreams, start=True):
    p = Pipe(istream, *ostreams)
    for ostream in ostreams:
        ostream.pipe_attached()

    if start:
        p.start()
    return p
